/*  
    This code is written by <albanese@fbk.it>.
    (C) 2010 Fondazione Bruno Kessler - Via Santa Croce 77, 38100 Trento, ITALY.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <Python.h>
#include <numpy/arrayobject.h>
#include <stdlib.h>
#include <math.h>


double
sgn (double x)
{
  if (x > 0.0)
    return 1.0;
  else if (x < 0.0)
    return -1.0;
  else
    return 0.0;
}


void matrix_dot_vector(double *A, double *b, double *out,
		       int nn, int pp)
{
  int n, p;
  
  for (n=0; n<nn; n++)
    {
      out[n] = 0.0;
      
      for (p=0; p<pp; p++)
	out[n] += A[p + (n * pp)] * b[p];  
    }
}


int stopping_rule(double *beta_curr, double *beta_new, int pp, double delta)
{
  int p;
  int ret = 1;
  
  for (p=0; p<pp; p++)
    {      
      if ( fabs(beta_new[p] - beta_curr[p]) > delta * fabs(beta_curr[p]) )
	{
	  ret = 0;
	  break;		       
	}
    }
  
  return ret;
}


/* precomputed omega = X^T Y */
void en_step(double *x, int nn, int pp, double *omega,
	     double tau, double mu, double C,
	     double *beta_curr, double *beta_new,
	     double *tmp)
{
  int n, p;
  double abs_b, half_tau, mu_plus_one;
 
  half_tau = tau * 0.5;  
  mu_plus_one = mu + 1.0;
 
  matrix_dot_vector(x, beta_curr, tmp, nn, pp);

  for (p=0; p<pp; p++)
    {
      /* beta_new = X^T tmp = X^T X beta_curr */
      beta_new[p] = 0.0;  
      for (n=0; n<nn; n++)
	beta_new[p] += x[p + (n * pp)] * tmp[n];  

      /* beta_new = beta_curr + (X^T Y - X^T X beta_curr) / C */
      beta_new[p] = beta_curr[p] + ((omega[p] - beta_new[p]) / C);

      /* beta_new = S_tau(beta_new) */
      abs_b = fabs(beta_new[p]);
      if (abs_b < half_tau)
	beta_new[p] = 0.0;
      else
	beta_new[p] = (abs_b - half_tau) * sgn(beta_new[p]); 

      /* beta_new =  beta_new / (1 + mu) */
      beta_new[p] = beta_new[p] / mu_plus_one;
    }
}


int en(double *x, int nn, int pp, double *omega,
       double tau, double mu, double C, double tol,
       int maxloops, double *beta_curr, double *beta_new)
{
  int i, ret;
  double *tmp, *tmp2;
  

  tmp = (double *) malloc (nn * sizeof(double));
    
  for (i=0; i<maxloops; i++)
    {
      en_step (x, nn, pp, omega, tau, mu, C, beta_curr, beta_new, tmp);
      ret = stopping_rule (beta_curr, beta_new, pp, tol / (double) (i + 1));
     
      tmp2 = beta_curr;
      beta_curr = beta_new;
      beta_new = tmp2;

      if ( ret == 1 )
	break;      
    }
 
  tmp2 = beta_curr;
  beta_curr = beta_new;
  beta_new = tmp2;

  free(tmp);

  return i;
}


static PyObject *elasticnet_elasticnet(PyObject *self, PyObject *args, PyObject *keywds)
{
  PyObject *x = NULL;
  PyObject *xC = NULL;
  double *xD;

  PyObject *omega = NULL;
  PyObject *omegaC = NULL;
  double *omegaD;

  PyObject *beta = NULL;
  PyObject *betaC = NULL;
  double *betaD;
  
  double tau, mu;
  double C, tol;
  int maxloops;

  PyObject *betaoutC = NULL;
  double *betaoutD;

  npy_intp nn, pp;
  npy_intp betaout_dims[1];

  double *beta_curr, *beta_new;
  int p, steps;


  /* Parse Tuple*/
  static char *kwlist[] = {"x", "omega", "tau", "mu", "beta", "C", "tol", "maxloops", NULL};
  if (!PyArg_ParseTupleAndKeywords(args, keywds, "OOddOddi", kwlist,
				   &x, &omega, &tau, &mu, &beta, &C, &tol, &maxloops))
    return NULL;
  
  xC = PyArray_FROM_OTF(x, NPY_DOUBLE, NPY_IN_ARRAY);
  if (xC == NULL) return NULL;

  omegaC = PyArray_FROM_OTF(omega, NPY_DOUBLE, NPY_IN_ARRAY);
  if (omegaC == NULL) return NULL;
  
  betaC = PyArray_FROM_OTF(beta, NPY_DOUBLE, NPY_IN_ARRAY);
  if (betaC == NULL) return NULL;

  if (PyArray_NDIM(xC) != 2)
    {
      PyErr_SetString(PyExc_ValueError, "x must be 2d array");
      return NULL;
    }

  if (PyArray_NDIM(omegaC) != 1)
    {
      PyErr_SetString(PyExc_ValueError, "omega must be 1d array");
      return NULL;
    }

  if (PyArray_NDIM(betaC) != 1)
    {
      PyErr_SetString(PyExc_ValueError, "beta must be 1d array");
      return NULL;
    }

  nn = PyArray_DIM(xC, 0);
  pp = PyArray_DIM(xC, 1);

  xD = (double *) PyArray_DATA(xC);
  omegaD = (double *) PyArray_DATA(omegaC);
  betaD = (double *) PyArray_DATA(betaC);
  
  beta_curr = (double *) malloc (pp * sizeof(double));
  beta_new = (double *) malloc (pp * sizeof(double));

  for (p=0; p<pp; p++)
    beta_curr[p] = betaD[p];

  steps = en(xD, nn, pp, omegaD, tau, mu, C, tol, maxloops, beta_curr, beta_new);

  betaout_dims[0] = pp;
  betaoutC = PyArray_SimpleNew (1, betaout_dims, NPY_DOUBLE);
  betaoutD = (double *) PyArray_DATA(betaoutC);

  for (p=0; p<pp; p++)
    betaoutD[p] = beta_new[p];

  Py_DECREF(xC);
  Py_DECREF(omegaC);
  Py_DECREF(betaC);

  free (beta_curr);
  free (beta_new);

  return Py_BuildValue("Ni", betaoutC, steps);
}


/* Doc strings: */
static char module_doc[] = "";

static char elasticnet_elasticnet_doc[] = "";

/* Method table */
static PyMethodDef elasticnet_methods[] = {
  {"elasticnet",
   (PyCFunction)elasticnet_elasticnet,
   METH_VARARGS | METH_KEYWORDS,
   elasticnet_elasticnet_doc},
  {NULL, NULL, 0, NULL}
};


/* Init */
void initelasticnet()
{
  Py_InitModule3("elasticnet", elasticnet_methods, module_doc);
  import_array();
}
