/*  
    This code is written by <albanese@fbk.it>.
    (C) 2010 Fondazione Bruno Kessler - Via Santa Croce 77, 38100 Trento, ITALY.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <Python.h>
#include <numpy/arrayobject.h>
#include <stdlib.h>
#include <math.h>
#include <float.h>


/* Compute exp( - sum( v_j^2 z_n(j) ) )
 * for each n.
 * 
 * See eq. 2.8 and 2.9 at page 533.
 */

void
exp_part(double *v, double *z, double *out, int nn, int pp)
{
  int n, p;

  for (n=0; n<nn; n++)
    {
      out[n] = 0.0;
      for (p=0; p<pp; p++)
	out[n] += pow(v[p], 2) * z[p + (n * pp)];
      
      out[n] = exp(-out[n]);
    }
}


/* See eq. 2.8 at page 533.
 *
 */

double
fun(double *ep, double *v, double lmbd, int nn, int pp)
{
  int n, p;
  double out = 0.0;
  double tmp = 0.0;
    
  for (p=0; p<pp; p++)
    tmp += pow(v[p], 2);
  tmp *= lmbd;

  for (n=0; n<nn; n++)
    out += log(1.0 + ep[n]);
  
  out += tmp;

  return out;
}


/* See eq. 2.9 at page 533.
 *
 */

void
grad_fun(double *ep, double *v, double *z, double lmbd,
	 double *out, int nn, int pp)
{
  int n, p;
  double tmp;


  for (p=0; p<pp; p++)
    out[p] = 0.0;

  for (n=0; n<nn; n++)
    {   
      tmp = ep[n] / (1.0 + ep[n]);
      
      for (p=0; p<pp; p++)
	out[p] += tmp * z[p + (n * pp)];   
    }
  
  for (p=0; p<pp; p++)
    out[p] = (lmbd - out[p]) * v[p]; 
}


/* Precompute g = c grad(f)^T grad(f)
 *
 * (for Backtracking Line Search).
 */

double
compute_g(double *grad, double c, int pp)
{
  int p;
  double g = 0.0;

  for (p=0; p<pp; p++)
    g += (grad[p] * grad[p]);
  
  g *= c;

  return g;
}


double
norm(double *x, int n)
{
  int i;
  double norm = 0.0;
  
  for (i=0; i<n; i++)
    norm += pow(x[i], 2);
  
  return sqrt(norm);
}


void
update_w(double *w, double *z, double lmbd, double eps,
	 double alpha0, double c, double rho,
	 double *w_out, int nn, int pp,
  	 double *v_buf, double *v_n_buf, double *grad_buf,
  	 double *ep_buf, double *ep_n_buf)
{

  int n, p;
  double A, B, alpha, g, nd;
 

  for (p=0; p<pp; p++)
    v_buf[p] = sqrt(w[p]);
  
  exp_part(v_buf, z, ep_buf, nn, pp);
  grad_fun(ep_buf, v_buf, z, lmbd, grad_buf, nn, pp);
  nd = DBL_MAX;
  
  while (nd >= eps)
    {      
      g = compute_g(grad_buf, c, pp);
      alpha = alpha0;
      
      /* v_n = v - alpha grad(f) */
      for (p=0; p<pp; p++)
	v_n_buf[p] = v_buf[p] - (alpha * grad_buf[p]);
	  
      /* A = f(v_n) */
      exp_part(v_n_buf, z, ep_n_buf, nn, pp);
      A = fun(ep_n_buf, v_n_buf, lmbd, nn, pp); 
      
      /* B = f(v) - alpha c grad(f)^T grad(f) */
      exp_part(v_buf, z, ep_buf, nn, pp);
      B = fun(ep_buf, v_buf, lmbd, nn, pp) - (alpha * g);

      while (!(A <= B))
	{  
	  alpha *= rho;
	  
	  /* v_n = v - alpha grad(f) */
	  for (p=0; p<pp; p++)
	    v_n_buf[p] = v_buf[p] - (alpha * grad_buf[p]);
	      
	  /* A = f(v_n) */
	  exp_part(v_n_buf, z, ep_n_buf, nn, pp);
	  A = fun(ep_n_buf, v_n_buf, lmbd, nn, pp); 
	  
	  /* B = f(v) - alpha c grad(f)^T grad(f) */
	  exp_part(v_buf, z, ep_buf, nn, pp);
	  B = fun(ep_buf, v_buf, lmbd, nn, pp) - (alpha * g);	  
	}
      
      for (n=0; n<nn; n++)
	ep_buf[n] = ep_n_buf[n];
	
      for (p=0; p<pp; p++)
	v_buf[p] = v_n_buf[p];
              
      grad_fun(ep_buf, v_buf, z, lmbd, grad_buf, nn, pp);
      nd = norm(grad_buf, pp);
    }
  
  for (p=0; p<pp; p++)
    w_out[p] = pow(v_n_buf[p], 2);
}


/* Compute matrix dk[i][j] = k(||x[i] - x[j]||_w).
 *
 * See step 3 in Figure 2 at page 534.
 */

void
compute_distance_kernel(double *x, double *w, int nn, int pp, 
			double sigma, double *out)
{
  int n1, n2, p;
  double d, dk;

  for (n1=0; n1<nn; n1++)
    for (n2=n1+1; n2<nn; n2++)
      {
	d = 0.0;
	for (p=0; p<pp; p++)
	  d += w[p] * fabs(x[p + (n1 * pp)] - x[p + (n2 * pp)]);
	
	dk = exp(-d / sigma);
	  
	out[n2 + (n1 * nn)] =  dk;
	out[n1 + (n2 * nn)] =  dk;
      }
}


static PyObject *
fssun_compute_distance_kernel(PyObject *self, PyObject *args, PyObject *keywds)
{
  PyObject *x = NULL;
  PyObject *xC = NULL;
  double *xD;

  PyObject *w = NULL;
  PyObject *wC = NULL;
  double *wD;

  double sigma;

  PyObject *outC = NULL;
  double *outD;
 
  npy_intp nn, pp;
  npy_intp out_dims[2];


  /* Parse Tuple*/
  static char *kwlist[] = {"x", "w", "sigma", NULL};
  if (!PyArg_ParseTupleAndKeywords(args, keywds, "OOd", kwlist,
				   &x, &w, &sigma))
    return NULL;
  
  xC = PyArray_FROM_OTF(x, NPY_DOUBLE, NPY_IN_ARRAY);
  if (xC == NULL) return NULL;

  wC = PyArray_FROM_OTF(w, NPY_DOUBLE, NPY_IN_ARRAY);
  if (wC == NULL) return NULL;
  
  nn = PyArray_DIM(xC, 0);
  pp = PyArray_DIM(xC, 1);

  xD = (double *) PyArray_DATA(xC);
  wD = (double *) PyArray_DATA(wC);
  
  out_dims[0] = nn;
  out_dims[1] = nn;

  outC = PyArray_SimpleNew (2, out_dims, NPY_DOUBLE);
  outD = (double *) PyArray_DATA (outC);

  compute_distance_kernel(xD, wD, (int) nn, (int) pp, 
			  sigma, outD);
  
  Py_DECREF(xC);
  Py_DECREF(wC);
  
  return Py_BuildValue("N", outC);
}


static PyObject *fssun_update_w(PyObject *self, PyObject *args, PyObject *keywds)
{
  PyObject *w = NULL;
  PyObject *wC = NULL;
  double *wD;

  PyObject *z = NULL;
  PyObject *zC = NULL;
  double *zD;

  PyObject *wnewC = NULL;
  double *wnewD;

  double lmbd, eps, alpha0, c, rho;
  
  npy_intp nn, pp;
  npy_intp wnew_dims[1];

  double *v_buf, *v_n_buf, *grad_buf, *ep_buf, *ep_n_buf;


  /* Parse Tuple*/
  static char *kwlist[] = {"w", "z", "lmbd", "eps", "alpha0", "c", "rho", NULL};
  if (!PyArg_ParseTupleAndKeywords(args, keywds, "OOddddd", kwlist,
				   &w, &z, &lmbd, &eps, &alpha0, &c, &rho))
    return NULL;
  
  wC = PyArray_FROM_OTF(w, NPY_DOUBLE, NPY_IN_ARRAY);
  if (wC == NULL) return NULL;

  zC = PyArray_FROM_OTF(z, NPY_DOUBLE, NPY_IN_ARRAY);
  if (zC == NULL) return NULL;
  
  nn = PyArray_DIM(zC, 0);
  pp = PyArray_DIM(zC, 1);

  wD = (double *) PyArray_DATA(wC);
  zD = (double *) PyArray_DATA(zC);
  
  wnew_dims[0] = pp;
  wnewC = PyArray_SimpleNew (1, wnew_dims, NPY_DOUBLE);
  wnewD = (double *) PyArray_DATA(wnewC);

  v_buf = (double *) malloc (pp * sizeof(double));
  v_n_buf = (double *) malloc (pp * sizeof(double));
  grad_buf = (double *) malloc (pp * sizeof(double));
  ep_buf = (double *) malloc (nn * sizeof(double));
  ep_n_buf = (double *) malloc (nn * sizeof(double));

  update_w(wD, zD, lmbd, eps, alpha0, c, rho, wnewD,
	   (int) nn, (int) pp, v_buf, v_n_buf, grad_buf,
	   ep_buf, ep_n_buf);

  free(v_buf);
  free(v_n_buf);
  free(grad_buf);
  free(ep_buf);
  free(ep_n_buf);
  
  Py_DECREF(wC);
  Py_DECREF(zC);
  
  return Py_BuildValue("N", wnewC);
}


/* Doc strings: */
static char module_doc[] = "";

static char fssun_update_w_doc[] = "";
static char fssun_compute_distance_kernel_doc[] = "";

/* Method table */
static PyMethodDef fssun_methods[] = {
  {"update_w",
   (PyCFunction)fssun_update_w,
   METH_VARARGS | METH_KEYWORDS,
   fssun_update_w_doc},
  {"compute_distance_kernel",
   (PyCFunction)fssun_compute_distance_kernel,
   METH_VARARGS | METH_KEYWORDS,
   fssun_compute_distance_kernel_doc},
  {NULL, NULL, 0, NULL}
};


/* Init */
void initfssun()
{
  Py_InitModule3("fssun", fssun_methods, module_doc);
  import_array();
}


 
