/*  
    This code is written by Davide Albanese <albanese@fbk.it>. 
    (C) 2010 Fondazione Bruno Kessler - Via Santa Croce 77, 38100 Trento, ITALY.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#include <Python.h>
#include <numpy/arrayobject.h>
#include <stdlib.h>
#include <math.h>


void
haar (double *x, double *y, int nn, double *scoeff,
      double *dcoeff, double *support, int levels)
{
  int n, j, shift;
  double lambda_n, lambda_shift;

  /* lambda_0 */
  support[0] = x[1] - x[0];
  support[nn-1] = x[nn-1] - x[nn-2];
  for (n=1; n<nn-1; n++)
    support[n] = ((x[n] + x[n+1]) / 2.0) - 
      ((x[n] + x[n-1]) / 2.0);

  /* scoeff_1 and dcoeff_1 */
  for (n=0; n<nn; n++)
    {
      shift = (n + 1) % nn;

      scoeff[n] = ((support[n] * y[n]) +
	(support[shift] * y[shift])) /
	(support[n] + support[shift]);
      
      dcoeff[n] = (sqrt(support[n] *
	support[shift]) * (y[shift] - y[n])) /
	(support[n] + support[shift]);		      
    }
  
  for (j=1; j<levels; j++)
    {
      /* compute lambda_j */
      for (n=0; n<nn; n++)
	{
	  shift = (n + (int) pow(2, j-1)) % nn;
	  
	  support[n + (j * nn)] = support[n + ((j-1) * nn)] + 
	    support[shift + ((j-1) * nn)];
	}

      /* scoeff_(j+1) and dcoeff__(j+1) */
      for (n=0; n<nn; n++)
	{
	  shift = (n + (int) pow(2, j)) % nn;

	  lambda_n = support[n + (j * nn)];
	  lambda_shift = support[shift + (j * nn)];

	  scoeff[n + (j * nn)] = ((lambda_n * scoeff[n + ((j-1) * nn)]) +
	    (lambda_shift *  scoeff[shift + ((j-1) * nn)])) /
	    (lambda_n + lambda_shift);
	  
	  dcoeff[n + (j * nn)] = sqrt(lambda_n * lambda_shift) * 
	    (scoeff[shift + ((j-1) * nn)] - scoeff[n + ((j-1) * nn)]) /
	    (lambda_n + lambda_shift);		      
	}
    }
}


static PyObject *irregularuwt_haar(PyObject *self, PyObject *args, PyObject *keywds)
{
  PyObject *x = NULL; PyObject *xa = NULL;
  PyObject *y = NULL; PyObject *ya = NULL;
  int levels = 0;
  
  double *_x, *_y;
  double *_S, *_D, *_Support;

  PyObject *Sa = NULL;
  PyObject *Da = NULL;
  PyObject *Supporta = NULL;
  
  npy_intp Sa_dims[2];
  npy_intp Da_dims[2];
  npy_intp Supporta_dims[2];

  int nn, J;


  /* Parse Tuple*/
  static char *kwlist[] = {"x", "y", "levels", NULL};
  if (!PyArg_ParseTupleAndKeywords(args, keywds, "OO|i",
				   kwlist, &x, &y, &levels))
    return NULL;

  xa = PyArray_FROM_OTF(x, NPY_DOUBLE, NPY_IN_ARRAY);
  if (xa == NULL) return NULL;

  ya = PyArray_FROM_OTF(y, NPY_DOUBLE, NPY_IN_ARRAY);
  if (ya == NULL) return NULL;
  
  nn = (int) PyArray_DIM(xa, 0);
  _x = (double *) PyArray_DATA(xa);
  _y = (double *) PyArray_DATA(ya);
   
  if (levels == 0)
    J = (int) log2(nn);
  else
    J = levels;

  Sa_dims[0] = (npy_intp) (J);
  Sa_dims[1] = PyArray_DIM(xa, 0);
  Sa = PyArray_SimpleNew(2, Sa_dims, NPY_DOUBLE);
  _S = (double *) PyArray_DATA(Sa);

  Da_dims[0] = (npy_intp) (J);
  Da_dims[1] = PyArray_DIM(xa, 0);
  Da = PyArray_SimpleNew(2, Da_dims, NPY_DOUBLE);
  _D = (double *) PyArray_DATA(Da);

  Supporta_dims[0] = (npy_intp) (J);
  Supporta_dims[1] = PyArray_DIM(xa, 0);
  Supporta = PyArray_SimpleNew(2, Supporta_dims, NPY_DOUBLE);
  _Support = (double *) PyArray_DATA(Supporta);

  haar (_x, _y, nn, _S, _D, _Support, J);
  
  Py_DECREF(xa);
  Py_DECREF(ya);

  return Py_BuildValue("N, N, N", Sa, Da, Supporta);
}


/* Doc strings: */
static char module_doc[]  = "";
static char irregularuwt_haar_doc[] = "";


/* Method table */
static PyMethodDef irregularuwt_methods[] = {
  {"haar",
   (PyCFunction)irregularuwt_haar,
   METH_VARARGS | METH_KEYWORDS,
   irregularuwt_haar_doc},
  {NULL, NULL, 0, NULL}
};

/* Init */
void initirregularuwt()
{
  Py_InitModule3("irregularuwt", irregularuwt_methods, module_doc);
  import_array();
}

