## This code is written by Davide Albanese, <albanese@fbk.eu>
## (C) 2010 Fondazione Bruno Kessler - Via Santa Croce 77, 38100 Trento, ITALY.

## LibLinear python class

## This program is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.

## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.

## You should have received a copy of the GNU General Public License
## along with this program.  If not, see <http://www.gnu.org/licenses/>.


import numpy as np
cimport numpy as np

cdef extern from "stdlib.h":
    ctypedef unsigned long size_t
    void free(void *ptr)
    void *malloc(size_t size)
    void *realloc(void *ptr, size_t size)
    size_t strlen(char *s)
    char *strcpy(char *dest, char *src)

cdef extern from "liblinear/linear.h":
    
    cdef struct feature_node:
        int index
        double value
    
    cdef struct problem:
        int l, n # n samples, n features
        int *y
        feature_node **x
        double bias
    
    cdef struct parameter:
        int solver_type
        double eps  # stopping criteria
        double C
        int nr_weight
        int *weight_label
        double* weight

    cdef struct model:
        parameter param	# parameter
        int nr_class	# number of classes
        int nr_feature
        double *w
        int *label    # label of each class
        double bias

    model* train(problem *prob, parameter *param)
    int predict(model *model_, feature_node *x)
    char *check_parameter(problem *prob, parameter *param)
    int save_model(char *model_file_name, model *model_)
    void free_and_destroy_model(model **model_ptr_ptr)
    void set_print_string_function(void (*print_func)(char *))

    
cdef void print_null(char *s):
    pass


cdef class LibLinear:
    cdef problem problem
    cdef parameter parameter
    cdef model *model 
    cdef feature_node *test_node
    cdef int p # number of features
               # -1 if no model computed
    
    SOLVER_TYPE = ['l2r_lr',
                   'l2r_l2loss_svc_dual',
                   'l2r_l2loss_svc',
                   'l2r_l1loss_svc_dual',
                   'mcsvm_cs',
                   'l1r_l2loss_svc',
                   'l1r_lr',
                   'l2r_lr_dual']

    def __cinit__(self, solver_type='l2r_lr', C=1, eps=0.01, weight={}):
        """LibLinear.
        
        :Parameters:        
            solver_type : string
                solver, can be one of 'l2r_lr', 'l2r_l2loss_svc_dual',
                'l2r_l2loss_svc', 'l2r_l1loss_svc_dual', 'mcsvm_cs',
                'l1r_l2loss_svc', 'l1r_lr', 'l2r_lr_dual'
            C : float
                cost of constraints violation
            eps : float
                stopping criterion
            weight : dict 
                changes the penalty for some classes (if the weight for a
                class is not changed, it is set to 1). For example, to
                change penalty for classes 1 and 2 to 0.5 and 0.8
                respectively set weight={1:0.5, 2:0.8}
        """


        try:
            self.parameter.solver_type = self.SOLVER_TYPE.index(solver_type)
        except ValueError:
            raise ValueError("invalid solver_type")

        self.parameter.C = C
        self.parameter.eps = eps
                
        # weight
        self.parameter.nr_weight = len(weight)
        self.parameter.weight_label = <int *> malloc \
            (len(weight) * sizeof(int))
        self.parameter.weight = <double *> malloc \
            (len(weight) * sizeof(double))
        try:
            for i, key in enumerate(weight):
                self.parameter.weight_label[i] = int(key)
                self.parameter.weight[i] = float(weight[key])
        except ValueError:
            raise ValueError("invalid weight")

        set_print_string_function(&print_null)
        self.p = -1


    def __dealloc__(self):
        self.__free_problem()
        self.__free_model()
        free(self.parameter.weight_label)
        free(self.parameter.weight)
        

    def __load_problem(self,
                       np.ndarray[np.float64_t, ndim=2] x not None,
                       np.ndarray[np.int64_t, ndim=1] y not None):
        """Convert the data into liblinear problem struct
        """
        

        cdef int i, j, k
        cdef np.ndarray[np.int64_t, ndim=1] nz

        if x.shape[0] != y.shape[0]:
            raise ValueError("x, y: shape mismatch")
        
        self.problem.x = <feature_node **> malloc \
            (x.shape[0] * sizeof(feature_node *))
        self.problem.y = <int *> malloc \
            (x.shape[0] * sizeof(double))

        for i in range(x.shape[0]):
            nz = np.nonzero(x[i])[0]
            self.problem.x[i] = <feature_node*> malloc \
                ((nz.shape[0]+1) * sizeof(feature_node))
            
            k = 0
            for j in nz:
                self.problem.x[i][k].index = j+1
                self.problem.x[i][k].value = x[i, j]
                k += 1

            self.problem.x[i][nz.shape[0]].index = -1
            self.problem.y[i] = y[i]
  
        self.problem.l = x.shape[0]
        self.problem.n = x.shape[1]
        self.problem.bias = -1
        

    def learn(self,
              np.ndarray[np.float64_t, ndim=2] x not None,
              np.ndarray[np.int64_t, ndim=1] y not None):
        """Constructs the model.
        
        :Parameters:
            x : 2d numpy array float
                training data (samples x features)
            y : 1d numpy array integer
                target values
        """


        cdef char *ret
        
        self.__free_problem()
        self.__free_model()
        self.p = -1

        self.__load_problem(x, y)
        ret = check_parameter(&self.problem, &self.parameter)
            
        if ret != NULL:
            raise ValueError(ret)
        
        self.model = train(&self.problem, &self.parameter)
        self.p = x.shape[1]


    def pred(self, np.ndarray[np.float64_t, ndim=2] x not None):
        """Does classification on test vectors x.
                
        :Parameters:
           x : 2d numpy array float
              test data (samples x features)
            
        :Returns:
           p : the predicted classes for x is returned.
        """


        cdef int i, j, k
        cdef np.ndarray[np.int64_t, ndim=1] p = \
            np.empty(x.shape[0], dtype=np.int64)
        cdef np.ndarray[np.int64_t, ndim=1] nz
        
        if self.p == -1:
            raise ValueError("no model computed")

        if x.shape[1] != self.p:
            raise ValueError("model, number of features: shape mismatch")        
        
        for i in range(x.shape[0]):
            nz = np.nonzero(x[i])[0]
            self.test_node = <feature_node*> malloc \
                ((nz.shape[0]+1) * sizeof(feature_node))
            
            k = 0
            for j in nz:
                self.test_node[k].index = j+1
                self.test_node[k].value = x[i, j]
                k += 1
                
            self.test_node[nz.shape[0]].index = -1
            p[i] = predict(self.model, self.test_node)
            free(self.test_node)
        
        return p


    def save_model(self, filename):
        """Saves a model to a file.        
        """
        
        cdef int ret
        
        if self.p == -1:
            raise ValueError("no model computed")
        
        ret = save_model(filename, self.model)
        
        if ret == -1:
            raise IOError("problem with save_model()")


    def get_nclasses(self):
        """Returns the number of classes.
        """
        
        if self.p == -1:
            raise ValueError("no model computed")

        return self.model.nr_class

    
    def weights(self):
        """Returns the feature weights.
        For 'mcsvm_cs' solver and for multiclass classification this
        method returns a dictionary label:w, where w is a 1d numpy
        array. For binary classification an 1d numpy array is
        returned.
        """
        
        cdef int i, n
        cdef np.ndarray[np.float64_t, ndim=1] w

        if self.p == -1:
            raise ValueError("no model computed")

        if (self.SOLVER_TYPE[self.parameter.solver_type] == \
                'mcsvm_cs') or self.model.nr_class != 2:
            n = self.problem.n * self.model.nr_class
        else:
            n = self.problem.n
    
        w = np.empty(n, dtype=np.float64)

        for i in range(n):
            w[i] = self.model.w[i]

        if (self.SOLVER_TYPE[self.parameter.solver_type] == \
                'mcsvm_cs') or self.model.nr_class != 2:
            w_res = w.reshape((-1, self.model.nr_class))
            w_dict = {}            
            for i in range(self.model.nr_class):
                w_dict[self.model.label[i]] = w_res[:, i]
            return w_dict
        else:
            return w


    cdef __free_problem(self):
        cdef int i
        
        if self.problem.x is not NULL:
            for i in range(self.problem.l):
                free(self.problem.x[i])
            free(self.problem.x)

        if self.problem.y is not NULL:
            free(self.problem.y)
        

    cdef __free_model(self):
        free_and_destroy_model(&self.model)
