## This code is written by Davide Albanese, <albanese@fbk.eu>
## (C) 2010 Fondazione Bruno Kessler - Via Santa Croce 77, 38100 Trento, ITALY.

## LibSVM python class

## This program is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.

## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.

## You should have received a copy of the GNU General Public License
## along with this program.  If not, see <http://www.gnu.org/licenses/>.


import numpy as np
cimport numpy as np

cdef extern from "stdlib.h":
    ctypedef unsigned long size_t
    void free(void *ptr)
    void *malloc(size_t size)
    void *realloc(void *ptr, size_t size)
    size_t strlen(char *s)
    char *strcpy(char *dest, char *src)

cdef extern from "libsvm/svm.h":
    
    cdef struct svm_node:
        int index
        double value
    
    cdef struct svm_problem:
        int l
        double *y
        svm_node **x
    
    cdef struct svm_parameter:
        int svm_type
        int kernel_type
        int degree	# for poly
        double gamma	# for poly/rbf/sigmoid
        double coef0	# for poly/sigmoid

        # these are for training only
        double cache_size # in MB
        double eps	# stopping criteria
        double C	# for C_SVC, EPSILON_SVR, and NU_SVR
        int nr_weight	      # for C_SVC
        int *weight_label     # for C_SVC
        double* weight	      # for C_SVC
        double nu	# for NU_SVC, ONE_CLASS, and NU_SVR
        double p	# for EPSILON_SVR
        int shrinking	# use the shrinking heuristics
        int probability # do probability estimates

    cdef struct svm_model:
        svm_parameter param	# parameter
        int nr_class	# number of classes, = 2 in regression/one class svm
        int l	        # total SV
        svm_node **SV	# SVs (SV[l])
        # coefficients for SVs in decision functions (sv_coef[k-1][l])
        double **sv_coef 
        double *rho     # constants in decision functions (rho[k*(k-1)/2])
        # pairwise probability information
        double *probA
        double *probB
        
        # for classification only
        int *label    # label of each class (label[k])
        int *nSV      # number of SVs for each class (nSV[k])
                      # nSV[0] + nSV[1] + ... + nSV[k-1] = l
        
        # XXX
        int free_sv  # 1 if svm_model is created by svm_load_model
                     # 0 if svm_model is created by svm_train

    void svm_set_print_string_function(void (*print_func)(char *))
    char *svm_check_parameter(svm_problem *prob, svm_parameter *param)
    svm_model *svm_train(svm_problem *prob, svm_parameter *param)
    double svm_predict(svm_model *model, svm_node *x)
    void svm_free_and_destroy_model(svm_model **model_ptr_ptr)
    int svm_save_model(char *model_file_name, svm_model *model)
    int svm_check_probability_model(svm_model *model)
    double svm_predict_probability(svm_model *model, svm_node *x,
                                   double* prob_estimates)
    
cdef void print_null(char *s):
    pass


cdef class LibSvm:
    cdef svm_problem problem
    cdef svm_parameter parameter
    cdef svm_model *model 
    cdef svm_node *test_node
    cdef int p # number of features
               # -1 if no model computed
    
    SVM_TYPE = ['c_svc',
                'nu_svc',
                'one_class',
                'epsilon_svr',
                'nu_svr']

    KERNEL_TYPE = ['linear',
                   'poly',
                   'rbf',
                   'sigmoid']

    def __cinit__(self, svm_type='c_svc', kernel_type='linear', 
                  degree=3, gamma=0.001, coef0=0, C=1, nu=0.5,
                  eps=0.001, p=0.1, cache_size=100, shrinking=True,
                  probability=False, weight={}):
        """LibSvm.
        
        :Parameters:
            svm_type : string
                SVM type, can be one of: 'c_svc', 'nu_svc', 
                'one_class', 'epsilon_svr', 'nu_svr'
            kernel_type : string
                kernel type, can be one of: 'linear' (u'*v),
                'poly' ((gamma*u'*v + coef0)^degree), 'rbf' 
                (exp(-gamma*|u-v|^2)), 'sigmoid'
                (tanh(gamma*u'*v + coef0))
            degree : int (for 'poly' kernel_type)
                degree in kernel
            gamma : float (for 'poly', 'rbf', 'sigmoid' kernel_type)
                gamma in kernel (e.g. 1 / number of features)
            coef0 : float (for 'poly', 'sigmoid' kernel_type)
                coef0 in kernel
            C : float (for 'c_svc', 'epsilon_svr', 'nu_svr')
                cost of constraints violation
            nu : float (for 'nu_svc', 'one_class', 'nu_svr')
                nu parameter
            eps : float
                stopping criterion, usually 0.00001 in nu-SVC,
                0.001 in others
            p : float (for 'epsilon_svr')
                p is the epsilon in epsilon-insensitive loss function
                of epsilon-SVM regression
            cache_size : float [MB]
                size of the kernel cache, specified in megabytes
            shrinking : bool
                use the shrinking heuristics
            probability : bool
                predict probability estimates
            weight : dict 
                changes the penalty for some classes (if the weight for a
                class is not changed, it is set to 1). For example, to
                change penalty for classes 1 and 2 to 0.5 and 0.8
                respectively set weight={1:0.5, 2:0.8}
        """


        try:
            self.parameter.svm_type = self.SVM_TYPE.index(svm_type)
        except ValueError:
            raise ValueError("invalid svm_type")

        try:
            self.parameter.kernel_type = self.KERNEL_TYPE.index(kernel_type)
        except ValueError:
            raise ValueError("invalid kerneltype")
        
        self.parameter.degree = degree
        self.parameter.gamma = gamma
        self.parameter.coef0 = coef0
        self.parameter.C = C
        self.parameter.nu = nu
        self.parameter.eps = eps
        self.parameter.p = p
        self.parameter.cache_size = cache_size
        self.parameter.shrinking = int(shrinking)
        self.parameter.probability = int(probability)
        
        # weight
        self.parameter.nr_weight = len(weight)
        self.parameter.weight_label = <int *> malloc \
            (len(weight) * sizeof(int))
        self.parameter.weight = <double *> malloc \
            (len(weight) * sizeof(double))
        try:
            for i, key in enumerate(weight):
                self.parameter.weight_label[i] = int(key)
                self.parameter.weight[i] = float(weight[key])
        except ValueError:
            raise ValueError("invalid weight")

        svm_set_print_string_function(&print_null)
        self.p = -1


    def __dealloc__(self):
        self.__free_problem()
        self.__free_model()
        free(self.parameter.weight_label)
        free(self.parameter.weight)
        

    def __load_problem(self,
                       np.ndarray[np.float64_t, ndim=2] x not None,
                       np.ndarray[np.float64_t, ndim=1] y not None):
        """Convert the data into libsvm svm_problem struct
        """
        

        cdef int i, j, k
        cdef np.ndarray[np.int64_t, ndim=1] nz

        if x.shape[0] != y.shape[0]:
            raise ValueError("x, y: shape mismatch")
        
        self.problem.x = <svm_node **> malloc \
            (x.shape[0] * sizeof(svm_node *))
        self.problem.y = <double *> malloc \
            (x.shape[0] * sizeof(double))

        for i in range(x.shape[0]):
            nz = np.nonzero(x[i])[0]
            self.problem.x[i] = <svm_node*> malloc \
                ((nz.shape[0]+1) * sizeof(svm_node))
            
            k = 0
            for j in nz:
                self.problem.x[i][k].index = j+1
                self.problem.x[i][k].value = x[i, j]
                k += 1

            self.problem.x[i][nz.shape[0]].index = -1
            self.problem.y[i] = y[i]
  
        self.problem.l = x.shape[0]
        
    def learn(self,
              np.ndarray[np.float64_t, ndim=2] x not None,
              np.ndarray[np.float64_t, ndim=1] y not None):
        """Constructs the model.
        
        :Parameters:
            x : 2d numpy array float
                training data (samples x features)
            y : 1d numpy array float
                target values
        """


        cdef char *ret
        
        self.__free_problem()
        self.__free_model()
        self.p = -1

        self.__load_problem(x, y)
        ret = svm_check_parameter(&self.problem, &self.parameter)
            
        if ret != NULL:
            raise ValueError(ret)
        
        self.model = svm_train(&self.problem, &self.parameter)
        self.p = x.shape[1]


    def pred(self, np.ndarray[np.float64_t, ndim=2] x not None):
        """Does classification or regression on test vectors x.
                
        :Parameters:
           x : 2d numpy array float
              test data (samples x features)
            
        :Returns:
           p : for a classification model, the predicted class for x is
              returned.  For a regression model, the function value of x
              calculated using the model is returned. For an one-class
              model, +1 or -1 is returned.
        """


        cdef int i, j, k
        cdef np.ndarray[np.float64_t, ndim=1] p = \
            np.empty(x.shape[0], dtype=np.float64)
        cdef np.ndarray[np.int64_t, ndim=1] nz

        if self.p == -1:
            raise ValueError("no model computed")

        if x.shape[1] != self.p:
            raise ValueError("model, number of features: shape mismatch")        
        for i in range(x.shape[0]):
            nz = np.nonzero(x[i])[0]
            self.test_node = <svm_node*> malloc \
                ((nz.shape[0]+1) * sizeof(svm_node))
            
            k = 0
            for j in nz:
                self.test_node[k].index = j+1
                self.test_node[k].value = x[i, j]
                k += 1

            self.test_node[nz.shape[0]].index = -1
            p[i] = svm_predict(self.model, self.test_node)
            free(self.test_node)
        
        return p

    
    def pred_probability(self, x):
        """Does classification or regression on a test vector x
        given a model with probability information.

        For a classification model with probability information, this
        method computes 'number of classes' probability estimates. The
        class with the highest probability is returned. For regression
        / one-class SVM, the returned value is the same as that of
        pred().
        
        :Parameters:
           x : 2d numpy array float
              test data (samples x features)
            
        :Returns:
           p, pe : for a classification model, 'p' is the predicted class
              for x. For a regression model, 'p' is the function value
              of x calculated using the model. For an one-class model,
              +1 or -1 is returned. 'pe' is a list of dictionaries
              containing probability estimates for each class.
        """
        

        cdef int i, j, k, ret
        cdef double *prob_estimates
        cdef np.ndarray[np.float64_t, ndim=1] p = \
            np.empty(x.shape[0], dtype=np.float64)
        cdef np.ndarray[np.int64_t, ndim=1] nz

        if self.p == -1:
            raise ValueError("no model computed")

        if x.shape[1] != self.p:
            raise ValueError("model, number of features: shape mismatch")        
        
        ret = svm_check_probability_model(self.model)
        if ret == 0:
            raise ValueError("model does not contain required information"
                             " to do probability estimates. Set probability=True")

        prob_estimates = <double*> malloc (self.model.nr_class * \
            sizeof(double))
        
        pe = []
        for i in range(x.shape[0]):
            nz = np.nonzero(x[i])[0]                
            self.test_node = <svm_node*> malloc \
                ((nz.shape[0]+1) * sizeof(svm_node))
            
            k = 0
            for j in nz:
                self.test_node[k].index = j+1
                self.test_node[k].value = x[i, j]
                k += 1

            self.test_node[nz.shape[0]].index = -1
            p[i] = svm_predict_probability(self.model,
                self.test_node, prob_estimates)
            free(self.test_node)

            if self.model.label is not NULL:
                pe.append(dict([(self.model.label[k],
                    prob_estimates[k]) for k in \
                    range(self.model.nr_class)]))        
        
        free(prob_estimates)
        
        return p, pe


    def save_model(self, filename):
        """Saves a model to a file.        
        """
        
        cdef int ret
        
        if self.p == -1:
            raise ValueError("no model computed")
        
        ret = svm_save_model(filename, self.model)
        
        if ret == -1:
            raise IOError("problem with svm_save_model()")


    def get_nclasses(self):
        """Get the number of classes.
        Returns 2 in regression and in one class SVM
        """
        
        if self.p == -1:
            raise ValueError("no model computed")

        return self.model.nr_class


    def get_nsv(self):
        """Get the number of support vectors.
        """
        
        if self.p == -1:
            raise ValueError("no model computed")

        return self.model.l

    
    def get_label_nsv(self):
        """Return a dictionary containing the number 
        of support vectors for each class (for classification).
        """

        cdef int i
        
        if self.p == -1:
            raise ValueError("no model computed")

        nsv = {}

        if self.model.nSV is not NULL:            
            for i in range(self.model.nr_class):
                nsv[self.model.label[i]] = self.model.nSV[i]
            
        return nsv

    
    cdef __free_problem(self):
        cdef int i
        
        if self.problem.x is not NULL:
            for i in range(self.problem.l):
                free(self.problem.x[i])
            free(self.problem.x)

        if self.problem.y is not NULL:
            free(self.problem.y)
        

    cdef __free_model(self):
        svm_free_and_destroy_model(&self.model)
