# mlpy setup file

from distutils.core import setup, Extension
from distutils.sysconfig import *
from distutils.util import *
from Cython.Distutils import build_ext
import os
import numpy

data_files = []

# Include gsl dlls for the win32 distribution
if get_platform() == "win32":
    dlls = ["mlpy\gslwin\libgsl-0.dll",
            "mlpy\gslwin\libgslcblas-0.dll"]
    data_files += [("Lib\site-packages\mlpy", dlls)]

## Extra compile args
extra_compile_args = ['-Wno-strict-prototypes']

# Python include
py_include = get_python_inc()

# NumPy include
numpy_lib = os.path.split(numpy.__file__)[0]
numpy_include = os.path.join(numpy_lib, 'core/include')

# NumPy support include
numpysupport_include = 'mlpy/numpysupport'

# Base include
base_include  = [py_include, numpy_include]

# Svm include
svm_include = base_include + [numpysupport_include,
                              'mlpy/svm/include']

# Knn include
knn_include = base_include + [numpysupport_include,
                             'mlpy/knn/include']

# Canberra include
canberra_include = base_include + [numpysupport_include]

# Svm sources
svm_sources = ['mlpy/svm/src/alloc.c', 'mlpy/svm/src/sort.c',
               'mlpy/svm/src/sampling.c', 'mlpy/svm/src/unique.c',
               'mlpy/svm/src/dist.c', 'mlpy/svm/src/svm.c',
               'mlpy/svm/src/matrix.c', 'mlpy/svm/src/rsfn.c',
               'mlpy/svm/src/rnd.c', 'mlpy/svm/svm.c',
               'mlpy/numpysupport/numpysupport.c']

# Knn sources
knn_sources = ['mlpy/knn/src/alloc.c', 'mlpy/knn/src/sort.c',
               'mlpy/knn/src/unique.c', 'mlpy/knn/src/dist.c',
               'mlpy/knn/src/nn.c', 'mlpy/knn/knn.c',
               'mlpy/numpysupport/numpysupport.c']

## Canberra sources
canberra_sources = ['mlpy/canberra/canberra.c',
                    'mlpy/numpysupport/numpysupport.c']


# Setup
setup(name = 'MLPY',
      version='2.2.2',
      requires=['numpy (>= 1.1.0)', 'gsl (>= 1.8)', 'cython (>= 0.13)'],
      description='mlpy - Machine Learning Py - ' \
          'High-Performance Python Package for Predictive Modeling',
      author='mlpy Developers - FBK-MPBA',
      author_email='albanese@fbk.eu',
      url='https://mlpy.fbk.eu',
      download_url='https://mlpy.fbk.eu/wiki/MlpyDownloads',
      license='GPLv3',
      classifiers=['Development Status :: 5 - Production/Stable',
                   'Intended Audience :: Science/Research',
                   'Intended Audience :: Developers',
                   'License :: OSI Approved :: GNU General Public License (GPL)',
                   'Natural Language :: English',
                   'Operating System :: POSIX :: Linux',
                   'Operating System :: POSIX :: BSD',
                   'Operating System :: Unix',
                   'Operating System :: MacOS :: MacOS X',
                   'Operating System :: Microsoft :: Windows',
                   'Programming Language :: C',
                   'Programming Language :: Python',
                   'Topic :: Scientific/Engineering :: Artificial Intelligence',
                   ],
      packages=['mlpy'],
      cmdclass = {'build_ext': build_ext},
      ext_modules=[Extension('mlpy.svm', svm_sources,
                             include_dirs=svm_include,
                             extra_compile_args=extra_compile_args),
                   Extension('mlpy.knn', knn_sources,
                             include_dirs=knn_include,
                             extra_compile_args=extra_compile_args),
                   Extension('mlpy.canberra', canberra_sources,
                             include_dirs=canberra_include,
                             extra_compile_args=extra_compile_args),
                   Extension('mlpy.hc', ['mlpy/hc/hc.c'],
                             include_dirs=base_include,
                             extra_compile_args=extra_compile_args),
                   Extension('mlpy.dwt', ['mlpy/dwt/dwt.c'],
                             include_dirs=base_include,
                             extra_compile_args=extra_compile_args,
                             libraries=['gsl', 'gslcblas', 'm']),
                   Extension('mlpy.uwt', ['mlpy/uwt/uwt.c'],
                             include_dirs=base_include,
                             extra_compile_args=extra_compile_args,
                             libraries=['gsl', 'gslcblas', 'm']),
                   Extension('mlpy.gslpy', ['mlpy/gslpy.c'],
                             include_dirs=base_include,
                             extra_compile_args=extra_compile_args,
                             libraries=['gsl', 'gslcblas', 'm']),
                   Extension('mlpy.cwb', ['mlpy/cwt/cwb.c'],
                             include_dirs=base_include,
                             extra_compile_args=extra_compile_args,
                             libraries=['gsl', 'gslcblas', 'm']),
                   Extension('mlpy.peaksd', ['mlpy/peaksd.c'],
                             include_dirs=base_include,
                             extra_compile_args=extra_compile_args),
                   Extension('mlpy.misc', ['mlpy/misc.c'],
                             include_dirs=base_include,
                             extra_compile_args=extra_compile_args),
                   Extension('mlpy.dtw', ['mlpy/dtw/dtw.c'],
                             include_dirs=base_include,
                             extra_compile_args=extra_compile_args),
                   Extension('mlpy.kmeans', ['mlpy/kmeans/kmeans.c'],
                             include_dirs=base_include,
                             extra_compile_args=extra_compile_args,
                             libraries=['gsl', 'gslcblas', 'm']),
                   Extension('mlpy.kernel', ['mlpy/kernel/kernel.c'],
                             include_dirs=base_include,
                             extra_compile_args=extra_compile_args,
                             libraries=['m']),
                   Extension('mlpy.spectralreg',
                             ['mlpy/spectralreg/spectralreg.c'],
                             include_dirs=base_include,
                             extra_compile_args=extra_compile_args),
                   Extension('mlpy.elasticnet',
                             ['mlpy/elasticnet/elasticnet.c'],
                             include_dirs=base_include,
                             extra_compile_args=extra_compile_args,
                             libraries=['m']),
                   Extension('mlpy.fssun',
                             ['mlpy/fssun/fssun.c'],
                             include_dirs=base_include,
                             extra_compile_args=extra_compile_args,
                             libraries=['m']),
                   Extension('mlpy.irregularuwt',
                             ['mlpy/iuwt/iuwt.c'],
                             include_dirs=base_include,
                             extra_compile_args=extra_compile_args),
                   Extension("mlpy.liblinear",
                             ["mlpy/liblinear/liblinear/linear.cpp", 
                              "mlpy/liblinear/liblinear/tron.cpp", 
                              "mlpy/liblinear/liblinear.pyx",
                              "mlpy/liblinear/liblinear/blas/daxpy.c",
                              "mlpy/liblinear/liblinear/blas/ddot.c",
                              "mlpy/liblinear/liblinear/blas/dnrm2.c",
                              "mlpy/liblinear/liblinear/blas/dscal.c"], 
                             extra_compile_args=extra_compile_args),
                   Extension("mlpy.libsvm",
                             ["mlpy/libsvm/libsvm/svm.cpp",
                              "mlpy/libsvm/libsvm.pyx"],
                             extra_compile_args=extra_compile_args)
                   ],
      scripts=["mlpy/tools/mlpy-test"],
      data_files=data_files
      )
